@@ -1,6 +1,6 @@
 (****************************************************************************)
 (* Copyright (C) 2007-2009 Gacek                                            *)
-(* Copyright (C) 2013-2016 Inria (Institut National de Recherche            *)
+(* Copyright (C) 2013-2022 Inria (Institut National de Recherche            *)
 (*                         en Informatique et en Automatique)               *)
 (*                                                                          *)
 (* This file is part of Abella.                                             *)
@@ -176,7 +176,7 @@ let pretty_obj ~left ~right ~printer obj =
         | [] -> concl
         | _ ->
             Opapp (10, Infix (LEFT, pretty_context ~printer obj.context,
-                              FMT " |-@ ", concl))
+                              FMT " |- ", concl))
       in
       Bracket {
         left ; right ; indent = 3 ; trans = OPAQUE ; inner
@@ -191,7 +191,7 @@ let pretty_obj ~left ~right ~printer obj =
         | _ -> Opapp (10, Infix (LEFT, pretty_context ~printer obj.context,
                                  FMT ",@ ", focus))
       in
-      let inner = Opapp (10, Infix (LEFT, hyps, FMT " |-@ ", concl)) in
+      let inner = Opapp (10, Infix (LEFT, hyps, FMT " |- ", concl)) in
       Bracket {
         left ; right ; indent = 3 ; trans = OPAQUE ; inner
       }
@@ -202,6 +202,40 @@ let pretty_obj obj =
     ~left:(STR "{") ~right:(STR "}")
     ~printer:Term.core_printer
 
+
+let pp_print_tids ff tids =
+  let open Format in
+  let pp_sep ff () = pp_print_space ff () in
+  if not !show_types then begin
+    pp_print_list ~pp_sep (fun ff (x, _) -> pp_print_string ff x) ff tids
+  end else begin
+    let rec collate cur_ids cur_ty res tids =
+      match tids with
+      | [] -> List.rev ((List.rev cur_ids, cur_ty) :: res)
+      | (x, ty) :: tids ->
+          if Term.eq_ty ty cur_ty then
+            collate (x :: cur_ids) cur_ty res tids
+          else
+            collate [x] ty ((List.rev cur_ids, cur_ty) :: res) tids
+    in
+    match tids with
+    | [] -> bugf "Empty quantified variables list"
+    | (x, ty) :: tids ->
+        let collated_tids = collate [x] ty [] tids in
+        let pp_collated ff (xs, ty) =
+          pp_open_box ff 2 ; begin
+            pp_print_string ff "(" ;
+            pp_print_list ~pp_sep pp_print_string ff xs ;
+            pp_print_string ff " : " ;
+            pp_print_string ff (ty_to_string ty) ;
+            pp_print_string ff ")"
+          end ; pp_close_box ff ()
+        in
+        pp_open_box ff 2 ; begin
+          pp_print_list ~pp_sep pp_collated ff collated_tids
+        end ; pp_close_box ff ()
+  end
+
 let rec pretty_metaterm mt =
   let open Format in
   match mt with
@@ -229,13 +263,7 @@ let rec pretty_metaterm mt =
           fun ff ->
             pp_print_string ff (binder_to_string q) ;
             pp_print_string ff " " ;
-            pp_open_box ff 0 ; begin
-              pp_print_string ff (fst (List.hd tids)) ;
-              List.iter begin fun tid ->
-                pp_print_space ff () ;
-                pp_print_string ff (fst tid)
-              end (List.tl tids) ;
-            end ; pp_close_box ff () ;
+            pp_print_tids ff tids ;
             pp_print_string ff ", " ;
         end) in
       let bod = pretty_metaterm a in
@@ -816,9 +844,9 @@ let rec meta_right_unify t1 t2 =
         meta_right_unify l1 l2 ;
         meta_right_unify r1 r2
     | Binding(b1, tids1, t1), Binding(b2, tids2, t2)
-        when b1 = b2 &&
-             List.length tids1 = List.length tids2 &&
-             List.for_all2 eq_ty (List.map snd tids1) (List.map snd tids2) ->
+      when b1 = b2 &&
+           List.length tids1 = List.length tids2 &&
+           tids_unifyable tids1 tids2 ->
         (* Replace bound variables with constants with "infinite"
            timestamp. This prevents illegal variable capture.
            We use max_int-1 since max_int is for nominal constants. *)
@@ -827,11 +855,23 @@ let rec meta_right_unify t1 t2 =
         in
         let alist1 = List.combine (List.map fst tids1) new_bindings in
         let alist2 = List.combine (List.map fst tids2) new_bindings in
-          meta_right_unify
-            (replace_metaterm_vars alist1 t1)
-            (replace_metaterm_vars alist2 t2)
+        meta_right_unify
+          (replace_metaterm_vars alist1 t1)
+          (replace_metaterm_vars alist2 t2)
     | _, _ -> raise (UnifyFailure Generic)
 
+and tids_unifyable tids1 tids2 =
+  let tys1 = List.map snd tids1 in
+  let tys2 = List.map snd tids2 in
+  let prob : Unifyty.constraints =
+    List.map2 (fun ty1 ty2 -> (ty1, ty2, Unifyty.(ghost, CArg))) tys1 tys2 in
+  let bstate = get_scoped_bind_state () in
+  match Unifyty.unify_constraints prob with
+  | _ -> true
+  | exception _ ->
+      set_scoped_bind_state bstate ;
+      false
+
 let try_meta_right_unify t1 t2 =
   try_with_state ~fail:false
     (fun () ->
