@@ -1,6 +1,6 @@
 (****************************************************************************)
 (* Copyright (C) 2007-2009 Gacek                                            *)
-(* Copyright (C) 2013-2018 Inria (Institut National de Recherche            *)
+(* Copyright (C) 2013-2022 Inria (Institut National de Recherche            *)
 (*                         en Informatique et en Automatique)               *)
 (*                                                                          *)
 (* This file is part of Abella.                                             *)
@@ -36,6 +36,9 @@ let lemmas : (string, string list * metaterm) H.t = State.table ()
 type subgoal = unit -> unit
 let subgoals : subgoal list ref = State.rref []
 
+let skip_seen : bool ref = State.rref false
+let start_proof () = skip_seen := false
+
 type hyp = {
   id : id ;
   term : metaterm ;
@@ -93,7 +96,7 @@ let close_types sign clauses atys =
   begin match List.intersect [oty; olistty; propty] tys with
   | [] -> ()
   | xs -> failwithf "Cannot close %s"
-     (String.concat ", " (List.map ty_to_string xs))
+            (String.concat ", " (List.map ty_to_string xs))
   end ;
   sr := Subordination.close !sr atys;
   List.iter (Typing.term_ensure_subordination !sr) (List.map snd clauses)
@@ -160,12 +163,6 @@ let clauses : clause list ref = State.rref []
 let add_clauses new_clauses =
   clauses := !clauses @ new_clauses
 
-let parse_defs ?(sign = !sign) str =
-  Lexing.from_string str |>
-  Parser.defs Lexer.token |>
-  type_udefs ~sr:!sr ~sign |>
-  List.map (fun (head, body) -> {head ; body})
-
 let defs_table : defs_table = State.table ()
 
 let built_ins_done = ref false
@@ -213,15 +210,15 @@ let register_definition = function
 
 let parse_definition str =
   Lexing.from_string str |>
-  Parser.top_command Lexer.token |>
-  register_definition
+  Parser.top_command_start Lexer.token |>
+  get_el |> register_definition
 
 let k_member = "member"
-let member_def_compiled =
-  "Define " ^ k_member ^ " : A -> list A -> prop by\
-  \  " ^ k_member ^ " A (A :: L) ; \
-  \  " ^ k_member ^ " A (B :: L) := " ^ k_member ^ " A L." |>
-  parse_definition
+let member_def_compiled = {|
+Define |} ^ k_member ^ {| : A -> list A -> prop by
+; |} ^ k_member ^ {| A (A :: L)
+; |} ^ k_member ^ {| A (B :: L) := |} ^ k_member ^ {| A L.
+|} |> parse_definition
 
 let () = built_ins_done := true
 
@@ -281,7 +278,7 @@ let instantiate_pred_types tysub (tbl: ty Itab.t) : ty Itab.t =
   Itab.iter begin fun id ty ->
     let ty' = apply_sub_ty tysub ty in
     res := Itab.add id ty' !res
-    end tbl;
+  end tbl;
   !res
 
 let instantiate_clause tysub cl =
@@ -304,14 +301,14 @@ let instantiate_clauses p def =
 let def_unfold term =
   match term with
   | Pred(p, _) ->
-     let pn = term_head_name p in
+      let pn = term_head_name p in
       if H.mem defs_table pn then begin
         let def = H.find defs_table pn in
         let (mutual, clauses) = instantiate_clauses p def in
         List.iter begin fun cl ->
           metaterm_ensure_subordination !sr cl.head ;
           metaterm_ensure_subordination !sr cl.body
-          end clauses;
+        end clauses;
         (mutual,clauses)
       end else
         failwith "Cannot perform case-analysis on undefined atom"
@@ -332,12 +329,12 @@ let separate_strings xs =
   let rec spin nxs pos = function
     | [] -> emit nxs
     | x :: xs ->
-      if pos + String.length x + 1 < __max_len then
-        spin (x :: nxs) (pos + String.length x + 1) xs
-      else begin
-        emit nxs ;
-        spin [x] (String.length x) xs
-      end
+        if pos + String.length x + 1 < __max_len then
+          spin (x :: nxs) (pos + String.length x + 1) xs
+        else begin
+          emit nxs ;
+          spin [x] (String.length x) xs
+        end
   in
   spin [] 0 xs ;
   List.rev !result
@@ -361,10 +358,8 @@ let format_vars ff =
     if !show_types then begin
       pp_print_newline ff () ;
       eigen_vars
-      |> List.map
-        (fun (x, xtm) -> (Term.tc [] xtm, x))
-      |> List.sort
-        (fun (ty1, _) (ty2, _) -> compare ty1 ty2)
+      |> List.map (fun (x, xtm) -> (Term.tc [] xtm, x))
+      |> List.sort (fun (ty1, _) (ty2, _) -> compare ty1 ty2)
       |> List.collate_assoc
       |> List.iter (format_typed_vars ff) ;
     end else begin
@@ -434,7 +429,7 @@ let set_subgoal_max ndps =
   let smax = !subgoal_max in
   let smax_map = List.fold_left begin
       fun smax_map (dp, n) ->
-        let n = Option.default max_int n in
+        let n = Option.value ~default:max_int n in
         (* Printf.printf "Max subgoals at depth %d: %d\n%!" dp n ; *)
         IntMap.add dp n smax_map
     end smax.smax_map ndps in
@@ -461,14 +456,17 @@ let format_other_subgoals fmt =
   format_count_subgoals fmt !count ;
   State.reload pristine
 
-let format_sequent fmt =
+let format_sequent_with_goal fmt goal =
   pp_open_vbox fmt 0 ; begin
     format_vars fmt ;
     format_hyps fmt ;
     fprintf fmt "============================@\n " ;
-    format_metaterm fmt sequent.goal
+    format_metaterm fmt goal
   end ; pp_close_box fmt ()
 
+let format_sequent fmt =
+  format_sequent_with_goal fmt sequent.goal
+
 let format_display fmt =
   pp_open_box fmt 0 ;
   if sequent.name = "" then
@@ -489,25 +487,47 @@ let get_display () =
   format_display (formatter_of_buffer b) ;
   Buffer.contents b
 
+let state_json () : Json.t =
+  let vars = List.map begin fun ((v, t) as vt) ->
+      `List [ `String v ;
+              if is_uninstantiated vt then `Null else
+                `String (term_to_string t) ]
+    end sequent.vars in
+  let hyps = List.map begin fun h ->
+      `List [
+        `String h.id ;
+        `String (metaterm_to_string h.term) ;
+      ]
+    end sequent.hyps in
+  let goal = metaterm_to_string sequent.goal in
+  `Assoc [
+    "more", `Int (List.length !subgoals) ;
+    "goal", `String goal ;
+    "hyps", `List hyps ;
+    "vars", `List vars ;
+    "name", `String sequent.name ;
+  ]
 
 (* Proof state manipulation utilities *)
 
-let reset_prover original_state original_sequent =
-  (* let original_state = get_bind_state () in
-   * let original_sequent = copy_sequent () in *)
-  fun () ->
-    set_bind_state original_state ;
-    set_sequent original_sequent ;
-    subgoals := []
+let reset_prover original_state original_sequent () =
+  set_bind_state original_state ;
+  set_sequent original_sequent ;
+  subgoals := []
 
 let full_reset_prover original_state original_sequent
-  original_clauses original_defs_table =
-  (* let original_clauses = !clauses in
-   * let original_defs_table = H.copy defs_table in *)
-  fun () ->
-    reset_prover original_state original_sequent () ;
-    clauses := original_clauses ;
-    H.assign defs_table original_defs_table
+    original_clauses original_defs_table () =
+  reset_prover original_state original_sequent () ;
+  clauses := original_clauses ;
+  H.assign defs_table original_defs_table
+
+(** Update sequent.count based on whether name is of the form
+    H<num> *)
+let update_hyp_count name =
+  if name.[0] <> 'H' then () else
+  match int_of_string (String.sub name 1 (String.length name - 1)) with
+  | n -> sequent.count <- max sequent.count n
+  | exception _ -> ()
 
 let add_hyp ?name term =
   let name = fresh_hyp_name begin
@@ -515,6 +535,7 @@ let add_hyp ?name term =
       | None -> ""
       | Some name -> name
     end in
+  update_hyp_count name ;
   sequent.hyps <- List.append sequent.hyps
       [{ id = name ; term = term ; abbrev = None }]
 
@@ -551,9 +572,9 @@ let add_if_new_var (name, v) =
     add_var (name, v)
 
 let add_lemma name tys lemma =
-  if H.mem lemmas name then
-    Format.eprintf "Warning: overriding existing lemma named %S@." name ;
-  H.replace lemmas name (tys, lemma)
+  let ovr = if H.mem lemmas name then `replace else `append in
+  H.replace lemmas name (tys, lemma) ;
+  ovr
 
 let get_hyp name =
   let hyp = List.find (fun h -> h.id = name) sequent.hyps in
@@ -604,11 +625,13 @@ let get_arg_clearly = function
   | Keep ("_", _) -> None
   | arg -> Some (get_stmt_clearly arg)
 
-exception End_proof of [`completed | `aborted]
+exception End_proof of [`completed of fin | `aborted]
 
 let next_subgoal () =
   match !subgoals with
-  | [] -> raise (End_proof `completed)
+  | [] ->
+      let fin = if !skip_seen then Unfinished else Finished in
+      raise (End_proof (`completed fin))
   | set_subgoal::rest ->
       set_subgoal () ;
       subgoals := rest ;
@@ -616,13 +639,12 @@ let next_subgoal () =
 
 (* Show *)
 
-let print_theorem name (tys, thm) =
-  let ff = Format.formatter_of_out_channel !Checks.out in
-  Format.fprintf ff "@[<hv2>Theorem %s%s :@ %a@].@."
+let print_theorem name (tys, thm) fmt =
+  Format.fprintf fmt "@[<hv2>Theorem %s%s :@ %a@].@."
     name (gen_to_string tys) format_metaterm thm
 
-let show name =
-  print_theorem name (get_generic_lemma name)
+let show name fmt =
+  print_theorem name (get_generic_lemma name) fmt
 
 (* Object level instantiation *)
 
@@ -700,7 +722,7 @@ let search_goal_witness ?depth goal witness =
              |> remove_coinductive_hypotheses
              |> List.map (fun h -> (h.id, h.term))
   in
-  let depth = Option.default !search_depth depth in
+  let depth = Option.value ~default:!search_depth depth in
   Tactics.search
     ~depth
     ~hyps
@@ -721,15 +743,12 @@ let search ?depth ~witness ~handle_witness () =
   | None -> failwith "Search failed"
   | Some w -> handle_witness w ; next_subgoal ()
 
-let async () =
-  failwith "async not implemented"
-
 (* Search cut *)
 
 let search_cut ?name h =
   match get_stmt_clearly h with
   | Obj(obj, _) ->
-      add_hyp ?name (Obj(search_cut ~search_goal obj, Irrelevant))
+      add_hyp ?name (Obj(Tactics.search_cut ~search_goal obj, Irrelevant))
   | _ ->
       failwith "Cut can only be used on hypotheses of the form {... |- ...}"
 
@@ -743,10 +762,13 @@ let goal_to_subgoal g =
     Term.set_bind_state bind_state ;
     sequent.goal <- g
 
-let ensure_no_logic_variable terms =
-  let logic_vars = List.flatten_map (metaterm_vars_alist Logic) terms in
+let ensure_no_logic_variable term =
+  let logic_vars = metaterm_vars_alist Logic term in
   if logic_vars <> [] then
-    failwith "Found logic variable at toplevel"
+    failwithf {|Found logic variable at toplevel.
+This usually means that the `search` tactic could not fully solve a goal.
+Note: the `search` tactic may have been invoked implicitly by `apply`,
+`backchain`, etc. with unknown arguments.|}
 
 let ensure_no_restrictions term =
   let rec aux t nested =
@@ -786,23 +808,32 @@ let type_apply_withs stmt ws =
     ws
 
 let partition_obligations ?depth obligations =
-  Either.partition_eithers
-    (List.map
-       (fun g -> match search_goal_witness ?depth g WMagic with
-          | None -> Either.Left g
-          | Some w -> Either.Right (g, w))
-       obligations)
+  List.partition_map begin fun g ->
+    (* Format.eprintf "before search:@.%a@." *)
+    (*   format_sequent_with_goal g ; *)
+    let wit = search_goal_witness ?depth g WMagic in
+    ensure_no_logic_variable g ;
+    match wit with
+    | None ->
+        (* Format.eprintf "after search/failure:@.%a@." *)
+        (*   format_sequent_with_goal g ; *)
+        Either.Left g
+    | Some w ->
+        (* Format.eprintf "after search/success:@.%a@." *)
+        (*   format_sequent_with_goal g ; *)
+        Either.Right (g, w)
+  end obligations
 
 let apply ?depth ?name ?(term_witness=ignore) h args ws =
   let stmt = get_stmt_clearly h in
   let args = List.map get_arg_clearly args in
-  let () = List.iter (Option.map_default ensure_no_restrictions ()) args in
+  let () = List.iter (Option.fold ~some:ensure_no_restrictions ~none:()) args in
   let ws = type_apply_withs stmt ws in
-  let result, obligations = Tactics.apply_with stmt args ws in
+  let result, obligations = Tactics.apply_with ~sr:!sr stmt args ws in
   let remaining_obligations, term_witnesses =
     partition_obligations ?depth obligations
   in
-  let () = ensure_no_logic_variable (result :: remaining_obligations) in
+  let () = ensure_no_logic_variable result in
   let () = List.iter term_witness term_witnesses in
   let obligation_subgoals = List.map goal_to_subgoal remaining_obligations in
   let resulting_case = recursive_metaterm_case ~used:sequent.vars ~sr:!sr result in
@@ -838,11 +869,10 @@ let type_backchain_withs stmt ws =
 let backchain ?depth ?(term_witness=ignore) h ws =
   let stmt = get_stmt_clearly h in
   let ws = type_backchain_withs stmt ws in
-  let obligations = Tactics.backchain_with stmt ws sequent.goal in
+  let obligations = Tactics.backchain_with ~sr:!sr stmt ws sequent.goal in
   let remaining_obligations, term_witnesses =
     partition_obligations ?depth obligations
   in
-  let () = ensure_no_logic_variable remaining_obligations in
   let () = List.iter term_witness term_witnesses in
   let obligation_subgoals = List.map goal_to_subgoal remaining_obligations in
   add_subgoals obligation_subgoals ;
@@ -982,7 +1012,7 @@ let coinduction ?name () =
 (* Assert *)
 
 let delay_mainline ?name ?depth new_hyp detour_goal =
-  if depth <> Some 0 && unwind_state (search_goal ?depth) detour_goal then
+  if depth <> Some 0 && search_goal ?depth detour_goal then
     add_hyp ?name new_hyp
   else
   let mainline =
@@ -1161,7 +1191,7 @@ let right () =
 let unfold clause_sel sol_sel =
   let mdefs = def_unfold sequent.goal in
   let used = sequent.vars in
-  let goal = unfold ~used ~mdefs clause_sel sol_sel sequent.goal in
+  let goal = unfold ~sr:!sr ~used ~mdefs clause_sel sol_sel sequent.goal in
   let goals = List.concat (List.map and_to_list goal) in
   add_subgoals (List.map goal_to_subgoal goals) ;
   next_subgoal ()
@@ -1196,6 +1226,7 @@ let exists ew =
 (* Skip *)
 
 let skip () =
+  skip_seen := true ;
   next_subgoal ()
 
 (* Clear *)
